# pylint: disable=no-self-use,unused-argument

"""Unit tests configuration file."""


import datafiles
import log
import pytest

from pomace import shared


class MockElement(str):
    @property
    def outer_html(self):
        return f"<mockhtml>{self}</>"

    @property
    def visible(self):
        return True


class MockLinks:
    def find_by_partial_text(self, value):
        return [MockElement(f"mockelement:links.partial_text={value}")]


class MockBrowser:

    url = "http://example.com"

    html = "Hello, world!"

    def find_by_name(self, value):
        return [MockElement(f"mockelement:name={value}")]

    def find_by_css(self, value):
        return [MockElement(f"mockelement:css={value}")]

    links = MockLinks()


@pytest.fixture
def mockbrowser(monkeypatch):
    browser = MockBrowser()
    monkeypatch.setattr(shared, "browser", browser)
    return browser


def pytest_configure(config):
    log.init(debug=True)
    log.silence("faker", "vcr")

    terminal = config.pluginmanager.getplugin("terminal")
    terminal.TerminalReporter.showfspath = False


def pytest_runtest_setup(item):
    datafiles.settings.HOOKS_ENABLED = False



"""Compatibility layer to handle systems that cannot run Playwright."""

# pylint: disable=unused-import
# mypy: ignore-errors

import sys

try:

    from playwright.sync_api import Browser as PlaywrightBrowser
    from playwright.sync_api import ElementHandle
    from playwright.sync_api import ElementHandle as PlaywrightElement
    from playwright.sync_api import Error as PlaywrightError
    from playwright.sync_api import Page
    from playwright.sync_api import sync_playwright as playwright

    if "pytest" in sys.modules:
        raise ImportError

except ImportError:

    class Missing:
        def __bool__(self):
            return False

    PlaywrightBrowser = Missing
    ElementHandle = Missing
    PlaywrightElement = Missing
    PlaywrightError = None
    Page = Missing
    playwright = Missing



import atexit
import inspect
import os
import sys
import time
from bdb import BdbQuit
from glob import iglob
from pathlib import Path

import ipdb
import log
from gitman.decorators import preserve_cwd
from gitman.models import Source
from selenium.common.exceptions import WebDriverException
from urllib3.exceptions import HTTPError

from . import browser, shared
from .config import settings


def launch_browser(
    delay: float = 0.0,
    *,
    silence_logging: bool = False,
    restore_previous_url: bool = True,
) -> bool:
    did_launch = False

    if silence_logging:
        log.silence("urllib3.connectionpool")

    if shared.browser:
        try:
            log.debug(f"Current browser windows: {shared.client.windows}")
            log.debug(f"Current browser URL: {shared.client.url}")
        except (WebDriverException, HTTPError) as e:
            log.warn(str(e).strip())
            shared.browser = None

    if not shared.browser:
        shared.browser = browser.launch()
        atexit.register(close_browser, silence_logging=silence_logging)
        did_launch = True

    if restore_previous_url and settings.url:
        shared.client.visit(settings.url, settings.browser.size)
        time.sleep(delay)

    return did_launch


def close_browser(*, silence_logging: bool = False) -> bool:
    did_close = False

    if silence_logging:
        log.silence("pomace", "selenium", allow_warning=True)

    if shared.browser:
        try:
            browser.save_url(shared.browser)
            browser.save_size(shared.browser)
            browser.close(shared.browser)
        except Exception as e:
            log.debug(e)
        else:
            did_close = True
        shared.browser = None

    path = Path().resolve()
    for _ in range(5):
        logfile = path / "geckodriver.log"
        if logfile.exists():
            log.debug(f"Deleting {logfile}")
            logfile.unlink()
        path = path.parent

    return did_close


def locate_models(*, caller=None):
    cwd = Path.cwd()
    default = cwd / "sites"

    if caller:
        for frame in inspect.getouterframes(caller):
            if "pomace" not in frame.filename or "pomace/tests" in frame.filename:
                path = Path(frame.filename)
                log.debug(f"Found caller's package directory: {path.parent}")
                os.chdir(path.parent)
                return

    if default.is_dir():
        log.info(f"Found models: {default}")
        return

    for name in iglob("./**/sites", recursive=True):
        path = Path(name).resolve()
        if path.is_dir():
            log.info(f"Found models: {path}")
            os.chdir(path.parent)
            return

    log.debug(f"Unable to locate models directory, using default: {default}")


@preserve_cwd
def clone_models(url: str, *, domain: str = "", force: bool = False):
    repository = url.replace(".git", "").split("/")[-1]
    domain = domain or repository.replace("pomace-", "")
    directory = Path("sites") / domain
    log.info(f"Cloning {url} to {directory}")
    if "." not in domain:
        raise ValueError(f"Invalid domain: {domain}")
    source = Source(url, directory)
    source.update_files(force=force)


def run_script(script: str):
    path = Path(script).resolve()
    if path.is_file():
        log.info(f"Running script: {path}")
    else:
        log.error(f"Script not found: {path}")
    try:
        exec(path.read_text())  # pylint: disable=exec-used
    except (KeyboardInterrupt, BdbQuit):
        pass
    except Exception as e:
        log.critical(f"Script exception: {e}")
        _type, _value, traceback = sys.exc_info()
        ipdb.post_mortem(traceback)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pomace', 'pomace.tests']

package_data = \
{'': ['*'], 'pomace.tests': ['cassettes/*']}

install_requires = \
['beautifulsoup4>=4.8.2,<5.0.0',
 'bullet>=2.2,<3.0',
 'cleo>=0.8.1,<0.9.0',
 'datafiles>=2.0,<3.0',
 'fake-useragent>=0.1.11,<0.2.0',
 'faker>=4.1.1',
 'flask-api>=3.0,<4.0',
 'gitman>=3.3,<4.0',
 'inflection>=0.5.1,<0.6.0',
 'ipdb>=0.13.9,<0.14.0',
 'ipython>=8.4,<9.0',
 'minilog>=2.0,<3.0',
 'nest-asyncio>=1.5.5,<2.0.0',
 'parse>=1.14,<2.0',
 'selenium>=4.4,<5.0',
 'splinter>=0.18.1,<0.19.0',
 'universal-startfile',
 'us>=2.0.2,<3.0.0',
 'webdriver_manager>=3.8.3,<4.0.0',
 'zipcodes>=1.1.2,<2.0.0']

extras_require = \
{':platform_machine != "armv7l"': ['playwright>=1.24,<2.0']}

entry_points = \
{'console_scripts': ['pomace = pomace.cli:application.run']}

setup_kwargs = {
    'name': 'pomace',
    'version': '0.11.7',
    'description': 'Dynamic page objects for browser automation.',
    'long_description': "# Pomace\n\nDynamic page objects for browser automation.\n\n[![Unix Build Status](https://img.shields.io/github/workflow/status/jacebrowning/pomace/main?label=unix)](https://github.com/jacebrowning/pomace/actions?query=branch%3Amain)\n[![Windows Build Status](https://img.shields.io/appveyor/ci/jacebrowning/pomace/main.svg?label=window)](https://ci.appveyor.com/project/jacebrowning/pomace)\n[![Coverage Status](https://img.shields.io/codecov/c/gh/jacebrowning/pomace)](https://codecov.io/gh/jacebrowning/pomace)\n[![PyPI Version](https://img.shields.io/pypi/v/pomace.svg)](https://pypi.org/project/pomace)\n[![PyPI License](https://img.shields.io/pypi/l/pomace.svg)](https://pypi.org/project/pomace)\n\n## Quick Start\n\nOpen **Terminal.app** in macOS and paste:\n\n```shell\npython3 -m pip install --upgrade pomace && python3 -m pomace run\n```\n\nor if you have Homebrew:\n\n```shell\nbrew install pipx; pipx run --no-cache pomace run\n```\n\n## Full Demo\n\nIf you're planning to run Pomace multiple times, install it with [pipx](https://pipxproject.github.io/pipx/) first:\n\n```shell\npipx install pomace\n```\n\nor get the latest version:\n\n```shell\npipx upgrade pomace\n```\n\nThen download some site models:\n\n```shell\npomace clone https://github.com/jacebrowning/pomace-twitter.com\n```\n\nAnd launch the application:\n\n```shell\npomace run twitter.com\n```\n\n# Usage\n\nInstall this library directly into an activated virtual environment:\n\n```shell\npip install pomace\n```\n\nor add it to your [Poetry](https://poetry.eustace.io/) project:\n\n```shell\npoetry add pomace\n```\n",
    'author': 'Jace Browning',
    'author_email': 'jacebrowning@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/pomace',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import platform
import subprocess
import sys

import log
import splinter
from fake_useragent import UserAgent, utils
from splinter.exceptions import DriverNotFoundError
from webdriver_manager import chrome, firefox

from . import patched
from .compat import PlaywrightError, playwright
from .config import settings
from .types import GenericBrowser, PlaywrightBrowser, SplinterBrowser

__all__ = ["launch", "save_url", "save_size", "close"]


NAMES = ["Firefox", "Chrome"]

PLAYWRIGHT_BROWSERS = {"chrome": "chromium"}

WEBDRIVER_MANAGERS = {
    "chromedriver": chrome.ChromeDriverManager,
    "geckodriver": firefox.GeckoDriverManager,
}

FALLBACK_USER_AGENT = "Mozilla/5.0 Gecko/20100101 Firefox/103.0"


def launch() -> GenericBrowser:
    if not settings.browser.name:
        log.error("No browser specified")
        sys.exit(1)

    if settings.browser.name == "open":
        settings.browser.name = NAMES[0]
    settings.framework = settings.framework.lower()
    settings.browser.name = settings.browser.name.lower()

    log.info(
        f"Launching {settings.browser.name.title()} using {settings.framework.title()}"
    )
    try:
        function = LAUNCHERS[settings.framework.title()]
    except KeyError:
        log.error(f"Unsupported framework: {settings.framework}")
        sys.exit(1)

    return function(settings.browser.name, settings.browser.headless)


def launch_playwright_browser(name: str, headless: bool) -> PlaywrightBrowser:
    name = PLAYWRIGHT_BROWSERS.get(name, name)
    _playwright = playwright().start()
    try:
        browser = getattr(_playwright, name)
    except AttributeError:
        log.error(f"Unsupported browser: {name}")
        sys.exit(1)

    try:
        instance = browser.launch(headless=headless)
    except PlaywrightError as e:
        if "playwright install" not in str(e):
            raise e from None  # pylint: disable=raising-bad-type
        subprocess.run((sys.executable, "-m", "playwright", "install", name))
        instance = browser.launch(headless=headless)

    setattr(instance, "_playwright", _playwright)
    return instance


def launch_splinter_browser(name: str, headless: bool) -> SplinterBrowser:
    utils.get_browsers = patched.get_browsers
    options = {
        "headless": headless,
        "user_agent": get_user_agent(name),
        "wait_time": 1.0,
    }
    log.debug(f"Options: {options}")
    try:
        return splinter.Browser(name, **options)
    except DriverNotFoundError:
        log.error(f"Unsupported browser: {name}")
        sys.exit(1)
    except Exception as e:  # pylint: disable=broad-except
        log.debug(str(e))

        if "exited process" in str(e):
            log.error("Browser update prevented launch. Please try again.")
            sys.exit(1)

        for driver, manager in WEBDRIVER_MANAGERS.items():
            if driver in str(e).lower():
                options["executable_path"] = manager().install()
                try:
                    return splinter.Browser(name, **options)
                except OSError as e:
                    if driver == "geckodriver" and "arm" in platform.machine():
                        log.error("Your machine's architecture is not supported")
                        log.info(
                            "https://firefox-source-docs.mozilla.org/testing/geckodriver/ARM.html"
                        )
                        sys.exit(1)
                    raise e from None

        raise e from None


LAUNCHERS = {
    "Playwright": launch_playwright_browser,
    "Splinter": launch_splinter_browser,
}
if playwright is None:
    del LAUNCHERS["Playwright"]


def get_user_agent(browser: str) -> str:
    user_agent = UserAgent(fallback=FALLBACK_USER_AGENT)
    return user_agent[browser]


def save_url(browser: GenericBrowser):
    if isinstance(browser, PlaywrightBrowser):
        page = browser.contexts[0].pages[0]
        url = page.url
    else:
        url = browser.url

    if "://" in url and settings.url != url:
        log.debug(f"Saving last browser URL: {url}")
        settings.url = url


def save_size(browser: GenericBrowser):
    if isinstance(browser, PlaywrightBrowser):
        page = browser.contexts[0].pages[0]
        # TODO: Figure out how to get the window size instead
        size: dict = page.viewport_size  # type: ignore
    else:
        size = browser.driver.get_window_size()

    if size != (settings.browser.width, settings.browser.height):
        log.debug(f"Saving last browser size: {size}")
        settings.browser.width = size["width"]
        settings.browser.height = size["height"]


def close(browser: GenericBrowser):
    if isinstance(browser, PlaywrightBrowser):
        browser.close()
        getattr(browser, "_playwright").stop()
    else:
        browser.quit()
